using System;
using System.IO;
using System.Text;


namespace Na.Sql
{
    public class Token : ICloneable
    {
        public string Def = "";
        public string Str = "";
        public string Type = "";

        #region ICloneable o

        public object Clone()
        {
            Token token = new Token();
            token.Def = Def;
            token.Str = Str;
            token.Type = Type;
            return token;
        }

        #endregion
    }

    public class Scanner
    {
        Token token;
        StreamReader input;
        int ungetChar;

        public Scanner(Stream input)
        {
            token = new Token();
            ungetChar = -1;
            this.input = new StreamReader(input, Encoding.Default);
        }

        public void Close()
        {
            input.Close();
        }

        bool IsEOF(int c)
        {
            return c == 0x1a || c <= -1;
        }

        bool IsWhilteSpace(int c)
        {
            return Char.IsWhiteSpace((char)c) && c != '\r' && c != '\n';
        }

        bool IsLetter(int c)
        {
            return Char.IsLetter((char)c);
        }

        bool IsDigit(int c)
        {
            return Char.IsDigit((char)c);
        }

        bool IsReserved(string str)
        {
            switch (str.ToUpper())
            {
                case "CREATE": return true;
                case "TABLE": return true;
                case "NOT": return true;
                case "NULL": return true;
                case "CONSTRAINT": return true;
                case "PRIMARY": return true;
                case "UNIQUE": return true;
                case "FOREIGN": return true;
                case "KEY": return true;
                case "REFERENCES": return true;
                case "ON": return true;
                case "DELETE": return true;
                case "UPDATE": return true;
                case "NO": return true;
                case "ACTION": return true;
                case "CASCADE": return true;
                case "SET": return true;
                case "DEFAULT": return true;

                case "CHAR": return true;
                case "VARCHAR": return true;
                case "CLOB": return true;
                case "NCHAR": return true;
                case "NCLOB": return true;
                case "BLOB": return true;
                case "INTEGER": return true;
                case "SMALLINT": return true;
                case "DECIMAL": return true;
                case "NUMERIC": return true;
                case "REAL": return true;
                case "FLOAT": return true;
                case "DOUBLE": return true;
                case "PRECISION": return true;
                case "DATE": return true;
                case "TIME": return true;
                case "TIMESTAMP": return true;
                case "INTERVAL": return true;
                    
                // for Oracle
                case "VARCHAR2": return true;
                case "NUMBER": return true;
                case "BYTE": return true;
            }

            return false;
        }

        public int GetCharacter()
        {
            int c = -1;
            if (ungetChar == -1)
            {
                c = input.Read();
            }
            else
            {
                c = ungetChar;
                ungetChar = -1;

            }
            return c;
        }

        public void UngetCharacter(int c)
        {
            if (ungetChar != -1)
                throw new InvalidOperationException();

            ungetChar = c;
        }

        public Token GetToken()
        {
            token.Str = "";
            token.Def = "EOF";
            token.Type = "EOF";

            int c = GetCharacter();
            while (IsWhilteSpace(c))
                c = GetCharacter();


            StringBuilder sb = new StringBuilder();
            if (IsLetter(c) || c == '_' || c == '"')
            {
                if (c == '"')
                    c = GetCharacter();
                
                // ʎq
                while (IsLetter(c) || IsDigit(c) || c == '_')
                {
                    sb.Append((char)c);
                    c = GetCharacter();
                }
                
                if (c != '"')
                    UngetCharacter(c);
                
                token.Str = sb.ToString();
                if (IsReserved(token.Str))
                {
                    token.Def = token.Str.ToUpper();
                }
                else
                {
                    token.Def = "IDENT";
                }
            }
            else if (IsDigit(c))
            {
                while (IsDigit(c))
                {
                    sb.Append((char)c);
                    c = GetCharacter();
                }

                if (c == '.')
                {
                    sb.Append((char)c);
                    while (IsDigit(c))
                    {
                        sb.Append((char)c);
                        c = GetCharacter();
                    }
                    UngetCharacter(c);
                    token.Str = sb.ToString();
                    token.Def = "DBL";
                }
                else
                {
                    UngetCharacter(c);
                    token.Str = sb.ToString();
                    token.Def = "NUM";
                }
            }
            else if (c == '.')
            {
                sb.Append((char)c);
                c = GetCharacter();
                if (IsDigit(c))
                {
                    while (IsDigit(c))
                    {
                        sb.Append((char)c);
                        c = GetCharacter();
                    }
                    UngetCharacter(c);
                    token.Str = sb.ToString();
                    token.Def = "DBL";
                }
                else
                {
                    UngetCharacter(c);
                    token.Str = ".";
                    token.Def = "PERIOD";
                }
            }
            else if (c == '\'')
            {
                c = GetCharacter();
                while ((char)c != '\'')
                {
                    sb.Append((char)c);
                    c = GetCharacter();
                }
                // '\''Ungetsv
                token.Str = sb.ToString();
                token.Def = "STR";

            }
            else if (c == '-')
            {
                sb.Append((char)c);
                c = GetCharacter();
                if (c == '-')
                {
                    // sRg
                    sb.Append((char)c);
                    token.Def = "LINECOMMENT";
                }
                else
                {
                    token.Def = "MINUS";
                    UngetCharacter(c);
                }
                token.Str = sb.ToString();
            }
            else if (c == '/')
            {
                sb.Append((char)c);
                c = GetCharacter();
                if (c == '*')
                {
                    // RgJn
                    sb.Append((char)c);
                    token.Def = "BEGINCOMMENT";
                }
                else
                {
                    UngetCharacter(c);
                    token.Def = "DIV";
                }
                token.Str = sb.ToString();
            }
            else if (c == '*')
            {
                sb.Append((char)c);
                c = GetCharacter();
                if (c == '/')
                {
                    // RgI
                    sb.Append((char)c);
                    token.Def = "ENDCOMMENT";
                }
                else
                {
                    UngetCharacter(c);
                    token.Def = "MUL";
                }
                token.Str = sb.ToString();
            }
            else if (c == ';')
            {
                token.Str = ";";
                token.Def = "SEMICOLON";
            }
            else if (c == ',')
            {
                token.Str = ",";
                token.Def = "COMMA";
            }
            else if (c == '(')
            {
                token.Str = "(";
                token.Def = "OPEN";
            }
            else if (c == ')')
            {
                token.Str = ")";
                token.Def = "CLOSE";
            }
            else if (c == '\r')
            {
                token.Str = "CR";
                token.Def = "CR";
            }
            else if (c == '\n')
            {
                token.Str = "LF";
                token.Def = "LF";
            }

            else if (c == -1)
            {
                token.Def = "EOF";
            }

            return token;
        }
    }
}